@@ -1,8 +1,8 @@
 package Perinci::Sub::ValidateArgs;
 
 # NOIFBUILT
-our $DATE = '2016-05-26'; # DATE
-our $VERSION = '0.009'; # VERSION
+our $DATE = '2016-06-02'; # DATE
+our $VERSION = '0.010'; # VERSION
 
 use 5.010001;
 use strict 'subs', 'vars';
@@ -65,9 +65,11 @@ sub gen_args_validator {
     my @meta_args = sort keys %$meta_args;
 
     my @code;
-    my %mod_stmts_cache;
+    my @modules_for_all_args;
     my @mod_stmts;
 
+    my $use_dpath;
+
     my $gencode_err = sub {
         my ($status, $term_msg) = @_;
         if ($args{die}) {
@@ -79,7 +81,7 @@ sub gen_args_validator {
             return "return [$status, $term_msg];";
         }
     };
-    my $gencode_validator = sub {
+    my $addcode_validator = sub {
         state $plc = do {
             require Data::Sah;
             Data::Sah->new->get_compiler("perl");
@@ -96,29 +98,25 @@ sub gen_args_validator {
                 return_type  => 'str',
                 indent_level => 2,
             );
+        die "Incompatible Data::Sah version (cd v=$cd->{v}, expected 2)" unless $cd->{v} == 2;
         #    $dsah_compile_cache{$cache_key} = $cd;
         #}
         push @code, "        \$err = undef;\n";
         push @code, "        \$_sahv_dpath = [];\n" if $cd->{use_dpath};
         push @code, "        unless (\n";
         push @code, $cd->{result}, ") { ".$gencode_err->(400, "\"Validation failed for argument '$data_name': \$err\"")." }\n";
-        for (@{ $cd->{modules} }) {
-            my $stmt;
-            if (my $ms = $cd->{module_statements}{$_}) {
-                $stmt = "$ms->[0] $_ (" . join(",", @{$ms->[1]}) . ");\n";
-            } else {
-                $stmt = "require $_;\n";
-            }
-            push @mod_stmts, $stmt unless $mod_stmts_cache{$stmt}++;
+        for my $mod_rec (@{ $cd->{modules} }) {
+            next unless $mod_rec->{phase} eq 'runtime';
+            next if grep { ($mod_rec->{use_statement} && $_->{use_statement} && $_->{use_statement} eq $mod_rec->{use_statement}) ||
+                               $_->{name} eq $mod_rec->{name} } @modules_for_all_args;
+            push @modules_for_all_args, $mod_rec;
+            push @mod_stmts, $plc->stmt_require_module($mod_rec)."\n";
+        }
+        if ($cd->{use_dpath}) {
+            $use_dpath = 1;
         }
     };
 
-    push @code, "sub {\n";
-    push @code, "    my \$args = shift;\n";
-    push @code, "    my \$err;\n";
-    push @code, "    my \$_sahv_dpath;\n";
-    push @code, "\n";
-
     if ($args_as eq 'hash' || $args_as eq 'hashref') {
         push @code, "    # check unknown args\n";
         push @code, "    for (keys %\$args) { unless (/\\A(".join("|", map { quotemeta } @meta_args).")\\z/) { ".$gencode_err->(400, '"Unknown argument \'$_\'"')." } }\n";
@@ -132,7 +130,7 @@ sub gen_args_validator {
                 push @code, "    $term_arg //= ".dmp($arg_spec->{default}).";\n";
             }
             push @code, "    if (exists $term_arg) {\n";
-            push @code, $gencode_validator->($arg_spec->{schema}, $arg_name, $term_arg) if $arg_spec->{schema};
+            $addcode_validator->($arg_spec->{schema}, $arg_name, $term_arg) if $arg_spec->{schema};
             if ($arg_spec->{req}) {
                 push @code, "    } else {\n";
                 push @code, "        ".$gencode_err->(400, "\"Missing required argument '$arg_name'\"")."\n";
@@ -151,8 +149,29 @@ sub gen_args_validator {
             push @code, "    # handle slurpy last arg\n";
             push @code, "    if (\@\$args >= $pos) { \$args->[$pos] = [splice \@\$args, $pos] }\n\n";
         }
-        push @code, "    # check extraneous args\n";
-        push @code, "    if (\@\$args > ".(@arg_names).") { ".$gencode_err->(400, "\"Too many arguments (expected ".(@arg_names).", got \".(\@\$args).\")\"") . " }\n";
+
+        my $start_of_optional;
+        for my $i (0..$#arg_names) {
+            my $arg_name = $arg_names[$i];
+            my $arg_spec = $meta_args->{$arg_name};
+            if ($arg_spec->{req}) {
+                if (defined $start_of_optional) {
+                    die "Error in metadata: after a param is optional ".
+                        "(#$start_of_optional) the rest (#$i) must also be optional";
+                }
+            } else {
+                $start_of_optional //= $i;
+            }
+        }
+
+        push @code, "    # check number of args\n";
+        if ($start_of_optional) {
+            push @code, "    if (\@\$args < $start_of_optional || \@\$args > ".(@arg_names).") { ".$gencode_err->(400, "\"Wrong number of arguments (expected $start_of_optional..".(@arg_names).", got \".(\@\$args).\")\"") . " }\n";
+        } elsif (defined $start_of_optional) {
+            push @code, "    if (\@\$args > ".(@arg_names).") { ".$gencode_err->(400, "\"Wrong number of arguments (expected 0..".(@arg_names).", got \".(\@\$args).\")\"") . " }\n";
+        } else {
+            push @code, "    if (\@\$args != ".(@arg_names).") { ".$gencode_err->(400, "\"Wrong number of arguments (expected ".(@arg_names).", got \".(\@\$args).\")\"") . " }\n";
+        }
         push @code, "\n";
 
         for my $i (0..$#arg_names) {
@@ -173,13 +192,15 @@ sub gen_args_validator {
             if (defined $arg_spec->{default}) {
                 push @code, "    $term_arg //= ".dmp($arg_spec->{default}).";\n";
             }
-            push @code, "    if (\@\$args > $i) {\n";
-            push @code, $gencode_validator->($arg_spec->{schema}, $arg_name, $term_arg) if $arg_spec->{schema};
-            if ($arg_spec->{req}) {
-                push @code, "    } else {\n";
-                push @code, "        ".$gencode_err->(400, "\"Missing required argument '$arg_name'\"")."\n";
+            my $open_block;
+            if (defined($start_of_optional) && $i >= $start_of_optional) {
+                $open_block++;
+                push @code, "    if (\@\$args > $i) {\n";
             }
-            push @code, "    }\n";
+            $addcode_validator->($arg_spec->{schema}, $arg_name, $term_arg) if $arg_spec->{schema};
+            push @code, "    }\n" if $open_block;
+
+            push @code, "\n";
         }
     } else {
         die "Unsupported args_as '$args_as'";
@@ -187,6 +208,14 @@ sub gen_args_validator {
     push @code, "    return undef;\n";
     push @code, "}\n";
 
+    unshift @code, (
+        "sub {\n",
+        "    my \$args = shift;\n",
+        "    my \$err;\n",
+        ("    my \$_sahv_dpath;\n") x !!$use_dpath,
+        "\n"
+    );
+
     my $code = join("", @mod_stmts, @code);
     if ($args{source}) {
         return $code;
@@ -213,7 +242,7 @@ Perinci::Sub::ValidateArgs - Validate function arguments using schemas in Rinci
 
 =head1 VERSION
 
-This document describes version 0.009 of Perinci::Sub::ValidateArgs (from Perl distribution Perinci-Sub-ValidateArgs), released on 2016-05-26.
+This document describes version 0.010 of Perinci::Sub::ValidateArgs (from Perl distribution Perinci-Sub-ValidateArgs), released on 2016-06-02.
 
 =head1 SYNOPSIS
 
